/*

 FILE

 fv_milter.c - FILE based vacation auto-responder and mail-forwarding agent,
 implemented as a Sendmail milter (mail filter) plugin, provides:

 * File based configuration
 * Multiple recipient support (cf. mailRoutingAddress)
 * Blacklisting
 * Catch-All support
 * Added SpamAssassin based Blacklist

 AUTHORS

 Thomas E. Horner <office@t-horner.com>
 ---
 Original version:
 Adam D. Gorski <adam@linuxbox.com>, Mike Gatny <mike@linuxbox.com>,
 Matt Benjamin <matt@linuxbox.com>


 COPYRIGHT
 
 Copyright (C) 2008 Thomas E. Horner
 Am Ipfbach 67
 A-4490 St. Florian, Austria
 http://www.t-horner.com
 ---
 Original version:
 Copyright (C) 2004,2005 The Linux Box Corp.
 206 S. Fifth Avenue Suite 150
 Ann Arbor, MI 48104
 http://www.linuxbox.com


 LICENSE

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the
 Free Software Foundation, Inc.,
 59 Temple Place - Suite 330, Boston, MA	02111-1307, USA.

*/

#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <libmilter/mfapi.h>
#include <time.h>
#include "strbuf.h"
#include "vacation.h"

#define VERSION "1.1"
#define MSG_FILE_BUFLEN 1024

/************************ Maximum length of a domain **************************/
#define MAX_DOMAIN_LENGTH 256


/****************************** Global varables *******************************/
int debug = 0;								/* debug mode, set with the -v flag */

int num_domains = 0;					// Number of domains
char **domains = NULL;				// Valid local domains

int num_blacklist = 0;					// Number of blacklist entries
char **blacklist = NULL;				// AutoRespond receiver blacklist

char *config_filename = NULL; /* set with -c option */


/**************************** Private memory struct ***************************/
typedef struct mlfiPriv {
	char	*mlfi_from;			/* Sender address */
	char	**mlfi_to;			/* Recepient address(es) */
	char	*subject;				/* Subject */
	int	 mlfi_to_count;		/* Number of recepients (internal counter) */
	int	 spam;						/* Spam message flag */
	int	 bulk;						/* Bulk message flag */
	char	*esmtpid;				/* EMSMTP ID */
} MlfiPriv;


/*************************** Config file struct *******************************/
struct configStruct {
	char *sendmail;							/* Sendmail executable path */
	char *domains_file;				 /* Local domains file */
	char *blacklist_file;				/* Blacklist file */
	char *msg_path;							/* Vacation message file path */
} config;


/********************* freemem - wrapper around free() ************************/
void
freemem (pointer, place)
	char *pointer;
	char *place;
{
	if (pointer != NULL) {
		free (pointer);
	} else {
		syslog(LOG_ERR, "Error: Tried to free NULL pointer: %s", place);
	}
}


/*********** read_domains - reads local domains from a file *******************/
int
read_domains(filename)
	char *filename;
{
	FILE *fp = NULL;						 // Domains file pointer
	char buf[MAX_DOMAIN_LENGTH]; // Domains file read buffer

	if ((fp = fopen(filename, "r")) != NULL) {
		while (fgets(buf, MAX_DOMAIN_LENGTH, fp) != NULL) {

			int i = strlen(buf);
			while (i>0)
			{
				i--;

				if (buf[i] == '\r' || buf[i] == '\n')
					buf[i] = '\0';
				else
					i = 0; //exit
			}
			if (buf[0] == '#' || buf[0] == ' ' || buf[0] == '\0') continue;


			if (debug) {
				syslog(LOG_INFO, "Debug: Adding %s to local domains.", buf);
			}
			
			domains = (char **) realloc((char **) domains, sizeof(char *) * (num_domains + 1));

			if (domains == NULL) {
				syslog(LOG_ERR, "Error: Failed to realloc() domain array.");
				fclose(fp);
				return 0;
			}

			domains[num_domains] = strdup(buf);

			if (domains[num_domains] == NULL) {
				syslog(LOG_ERR, "Error: Failed to strdup() domain.");
				fclose(fp);
				return 0;
			}

			strncpy(domains[num_domains], buf, strlen(buf));
			num_domains++;
		}

		fclose(fp);
	} else {
		fprintf(stderr, "Could not open %s for reading\n", filename);
		exit(EX_SOFTWARE);
	}

	return 1;
}


/*********** read_blacklist - reads autorespond-receiver blacklist from a file *******************/
int
read_blacklist(filename)
	char *filename;
{
	FILE *fp = NULL;						 // Blacklist file pointer
	char buf[MAX_DOMAIN_LENGTH]; // Blacklist file read buffer

	if ((fp = fopen(filename, "r")) != NULL) {
		while (fgets(buf, MAX_DOMAIN_LENGTH, fp) != NULL) {

			int i = strlen(buf);
			while (i>0)
			{
				i--;

				if (buf[i] == '\r' || buf[i] == '\n')
					buf[i] = '\0';
				else
					i = 0; //exit
			}
			if (buf[0] == '#' || buf[0] == ' ' || buf[0] == '\0') continue;


			if (debug) {
				syslog(LOG_INFO, "Debug: Adding %s to blacklist.", buf);
			}
			
			blacklist = (char **) realloc((char **) blacklist, sizeof(char *) * (num_blacklist + 1));

			if (blacklist == NULL) {
				syslog(LOG_ERR, "Error: Failed to realloc() blacklist array.");
				fclose(fp);
				return 0;
			}

			blacklist[num_blacklist] = strdup(buf);

			if (blacklist[num_blacklist] == NULL) {
				syslog(LOG_ERR, "Error: Failed to strdup() blacklist entry.");
				fclose(fp);
				return 0;
			}

			strncpy(blacklist[num_blacklist], buf, strlen(buf));
			num_blacklist++;
		}

		fclose(fp);
	} else {
		fprintf(stderr, "Could not open %s for reading\n", filename);
		exit(EX_SOFTWARE);
	}

	return 1;
}


/*************** rtrimBracket - trims > from an e-mail address ****************/
void
rtrimBracket(string)
	char *string;
{
	int i = strlen(string) - 1; /* String index */

	while(i >= 0 && string[i] == '>') {
		string[i--] = '\0';
	}
} /* rtrimBracket */


/******* local_domain - checks domain of receiver against local domains *******/
int
local_domain(email)
	char *email;
{
	char *tmp = NULL; // Temporary character location pointer
	int size,				 	// String size
	i;								// Domain array index
	
	if ((tmp = strchr(email, '@')) == NULL) {
		// We did not find an @, might be a bare username
		if (strlen(email)) {
			return 1;
		} else {
			return 0;
		}
	}

	tmp++;
	rtrimBracket(tmp);

	if (domains != NULL) {
		for (i = 0; i < num_domains; i++) {
			size = strlen(domains[i]);

			if (!strncasecmp(tmp, domains[i], size)) {
				return 1;
			}
		}
	}

	return 0;
} // local_domain


/******* is_blacklisted - checks email adress of receiver against blacklist *******/
int
is_blacklisted(email)
	char *email;
{
	int size,				 	// String size
	i;								// Blacklist array index

	if (blacklist != NULL) {
		for (i = 0; i < num_blacklist; i++) {
			size = strlen(blacklist[i]);

			if (!strncasecmp(email, blacklist[i], size)) {

				if (debug) {
					syslog(LOG_INFO, "Debug: %s found to be blacklisted", email);
				}

				return 1;
			}
		}
	}

	if (debug) {
		syslog(LOG_INFO, "Debug: %s found to NOT be blacklisted", email);
	}

	return 0;
} // is_blacklisted


/************************** mlfi_message_cleanup ******************************/
sfsistat
mlfi_message_cleanup(ctx)
	SMFICTX *ctx;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);
	int i; /* "To:" array index */

	/* Clean up "per message" resources */
	if (priv != NULL) {
		if (debug) {
			syslog(LOG_INFO, "Debug: Message Cleanup!");
		}

		if (priv->mlfi_from != NULL) {
			freemem(priv->mlfi_from, "priv->mlfi_from");
			priv->mlfi_from = NULL;
		}

		if (priv->mlfi_to != NULL) {
			for (i = 0; i < priv->mlfi_to_count; i++) {
				freemem(priv->mlfi_to[i], "priv->mlfi_to[i]");
			}

			if (priv->mlfi_to != NULL) {
				freemem(priv->mlfi_to, "priv->mlfi_to");
				priv->mlfi_to = NULL;
			}
		}

		if (priv->subject != NULL) {
			freemem(priv->subject, "priv->subject");
			priv->subject = NULL;
		}

		priv->mlfi_to_count = 0;
		priv->spam = 0;
		priv->bulk = 0;
	}

	return SMFIS_CONTINUE;
} /* mlfi_message_cleanup */


/************************ mlfi_connection_cleanup *****************************/
sfsistat
mlfi_connection_cleanup(ctx)
	SMFICTX *ctx;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);

	/* Clean up "per connection" resources */
	if (priv != NULL) {
		if (debug) {
			syslog(LOG_INFO, "Debug: Connection Cleanup!");
		}

		free(priv->esmtpid);
		freemem(priv, "priv");

		smfi_setpriv(ctx, NULL);
	}

	return SMFIS_CONTINUE;
} /* mlfi_connection_cleanup */



/***************************** extract_account ******************************/
char *
extract_account(email)
	char *email;
{
	char *tmp = NULL;			/* Temporary character location pointer */
	char *username = NULL; /* Username string pointer */
	int size = 0;					/* String size */
	int strindex = 0;			/* String index */
	int length = 0;				/* String length */
	
	int atpos = 0;

	length = strlen(email);

	/* Find the @ sign in the e-mail */
	if ((tmp = strchr(email, '@')) == NULL) {
		/* We did not find an @, might be a bare username */
		strindex = length;
	} else {
		atpos = length - strlen(tmp);
		strindex = atpos;
	}

	/* work to the left until we hit whitespace or the left angle bracket */
	while ((strindex != -1) && !isspace(email[strindex]) && (email[strindex] != '<')) {
		strindex--;
		size++;
	}

	if (tmp == NULL) {
		/* we're now at the left edge */
		size--;
		tmp = email + strindex + 1;
	
		username = (char *) malloc (sizeof(char) * (size + 1));
		strncpy(username, tmp, size);
		username[size] = '\0';
	
		if(username[strlen(username) - 1] == '>') {
			username[strlen(username) - 1] = '\0';
		}
	} else {
		//first we copy the domain
		int endindex = atpos;
		while ((endindex <= length) && !isspace(email[endindex]) && (email[endindex] != '>')) {
			endindex++;
		}
		
		username = (char *) malloc (sizeof(char) * (size + 1 + endindex - atpos));
		strncpy(username, email + atpos + 1, endindex - atpos - 1);
		username[endindex - atpos - 1] = '#';
		
		tmp = email + strindex + 1;
		strncpy(username + endindex - atpos, tmp, size - 1);
		username[size + endindex - atpos - 1] = '\0';
	
		if(username[strlen(username) - 1] == '>') {
			username[strlen(username) - 1] = '\0';
		}
	}

		return username;
} /* extract_account */



/**************************** extract_domainname *****************************/
char *
extract_domainname(email)
	char *email;
{
	char *tmp = NULL;			/* Temporary character location pointer */
	char *username = NULL; /* Username string pointer */
	int size = 0;					/* String size */
	int strindex = 0;			/* String index */
	int length = 0;				/* String length */

	length = strlen(email);

	/* Find the @ sign in the e-mail */
	if ((tmp = strchr(email, '@')) == NULL) {
		/* We did not find an @, might be a bare username */
		strindex = length;
	} else {
		strindex = length - strlen(tmp);
	}

	/* work to the right until we hit whitespace or the right angle bracket */
	while ((strindex < length) && !isspace(email[strindex]) && (email[strindex] != '>')) {
		strindex++;
		size++;
	}

	/* we're now at the right edge */
	size--;
	tmp++;

	username = (char *) malloc (sizeof(char) * (size + 1));
	strncpy(username, tmp, size);
	username[size] = '\0';

	if(username[strlen(username) - 1] == '>') {
		username[strlen(username) - 1] = '\0';
	}

		return username;
} /* extract_domainname */



/****************** save - saves config option from file	*********************/
int
save (buf, conf)
	char *buf;
	char **conf;
{
	if (!sscanf(buf, "%*s%*[ \t]%[^\n]", buf)) {
		fprintf(stderr, "Invalid configuration line: %s\n", buf);
		return 0;
	}

	*conf = strdup(buf);

	if (*conf == NULL) {
		fprintf(stderr, "Failed to strdup() conf value in save()\n");
		return 0;
	}

	return 1;
}


/******************* read_config - reads config file	*************************/
int
read_config ()
{
	FILE *fp = NULL;	/* Configuration file pointer */
	int len = 0,			/* String length */
			error = 0;		/* Local error flag */
	char *pos = NULL, /* File position */
			 buf[1024];	 /* File read buffer */

	if (debug) {
		fprintf(stderr, "Reading config file:\n\n");
	}

	/* if a config file was specified on the command line, use it */	
	if (config_filename != NULL) {
		fp = fopen(config_filename, "r");

		if(!fp) {
			fprintf(stderr, "Specified config file failed to open, defaulting to /etc/fv_milter.conf\n");
		}
	}

	/* if a config file was not specified, or the specified file failed to open,
	 * default to /etc/fv_milter.conf
	 */ 
	if(!fp) {
		fp = fopen( "/etc/fv_milter.conf", "r" );
	}

	if (!fp) {
		fprintf(stderr,	"/etc/fv_milter.conf failed to open. A valid configuration file must be present for the milter to run.\n");
		return 0;
	}

	while (fgets(buf, 1024, fp)) { 
		if (strlen(buf) == 1) {
			continue; /* skip empty lines */
		}

		if ((pos = strchr(buf, '#'))) {
			if (strlen(pos) == strlen(buf)) {
				continue; /* comment at the beginning, skip */
			} else {
				/* Extract the option before the comment */
				len = strlen(buf) - strlen(pos) - 1;

				*pos = '\0';

				while ((len > 0) && isspace(buf[len])) {
					buf[len] = '\0';
					len--;
				}

				buf[++len] = '\n';
			}
		}
		 
		if (debug) {
			fprintf(stderr, "\t%s", buf);
		}

		if (strstr(buf, "SENDMAIL")) {
			if (!save(buf, &config.sendmail)) {
				fclose(fp);
				return 0;
			}
		} else if (strstr(buf, "DOMAINS_FILE")) {
			if (!save(buf, &config.domains_file)) {
				fclose(fp);
				return 0;
			}
		} else if (strstr(buf, "BLACKLIST_FILE")) {
			if (!save(buf, &config.blacklist_file)) {
				fclose(fp);
				return 0;
			}
		} else if (strstr(buf, "MSG_PATH")) {
			if (!save(buf, &config.msg_path)) {
				fclose(fp);
				return 0;
			}
		}
	}

	if (config.sendmail == NULL) {
		fprintf(stderr, "SENDMAIL configuration option invalid or missing\n");
		error = 1;
	} else if (config.domains_file == NULL) {
		fprintf(stderr, "DOMAINS_FILE configuration option invalid or missing\n");
		error = 1;
	} else if (config.blacklist_file == NULL) {
		fprintf(stderr, "Optional BLACKLIST_FILE configuration option invalid or missing\n");
		//NOT AN ERROR
	} else if (config.msg_path == NULL) {
		fprintf(stderr, "MSG_PATH configuration option invalid or missing\n");
		error = 1;
	}

	fclose(fp);

	return !error;
}


/******************* usage - prints usage information *************************/
void
usage (prog)
	char *prog;
{
		fprintf(stderr,
		"fv_milter: Vacation milter with FILE based configuration\n\n"
		"Valid options\n"
		"	-p [local:|unix:][socket_path]	Specifies the unix socket to listen on\n"
		"	-c [config_file_path]					 Specifies a non-standard config file location\n"
		"	-d															Turns on debugging mode\n"
		"	-v															Prints out version number\n\n"
		"	Example: %s -p local:/var/run/fv_milter.sock\n\n", prog);
} 


/******************************* mlfi_connect *********************************/
sfsistat
mlfi_connect(ctx, hostname, hostaddr)
	SMFICTX *ctx;
	char *hostname;
	_SOCK_ADDR *hostaddr;
{
	MlfiPriv *priv; /* Private memory struct pointer */

	if (debug) {
		syslog(LOG_INFO, "Debug: New connection from %s", hostname);
	}

	/* Allocate private memory */
	priv = malloc(sizeof *priv);

	if (priv == NULL) {
		syslog(LOG_ERR, "Error: Failed to malloc() private memory in mlfi_connect()");
		return SMFIS_TEMPFAIL;
	}

	memset(priv, '\0', sizeof *priv);

	/* Save the private memory context struct */
	smfi_setpriv(ctx, priv);

	/* NULL out all pointers */
	priv->mlfi_from = NULL;
	priv->mlfi_to = NULL;
	priv->subject = NULL;

	/* Zero out all ints */
	priv->mlfi_to_count = 0;
	priv->spam = 0;
	priv->bulk = 0;

	return SMFIS_CONTINUE;
} /* mlfi_connect */


/****************************** mlfi_envfrom **********************************/
sfsistat
mlfi_envfrom(ctx, argv)
	SMFICTX *ctx;
	char **argv;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);

	if (debug) {
		syslog(LOG_INFO, "Debug: New message");
	}

	priv->mlfi_from = strdup(argv[0]);

	if (priv->mlfi_from == NULL) {
		syslog(LOG_ERR, "Error: Failed to strdup() from address in mlfi_envfrom()");
		(void) mlfi_message_cleanup(ctx);
		return SMFIS_TEMPFAIL;
	}

	priv->esmtpid = strdup(smfi_getsymval(ctx, "i"));
	
	if (debug) {
		syslog(LOG_INFO, "Debug: From: %s", priv->mlfi_from);
	}

	return SMFIS_CONTINUE;
} /* mlfi_envfrom */


/****************************** mlfi_envrcpt **********************************/
sfsistat
mlfi_envrcpt(ctx, argv)
	SMFICTX *ctx;
	char **argv;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);
	char *tempto; // Temporary "To:" address pointer

	if (debug) {
		syslog(LOG_INFO, "Debug: To: %s", argv[0]);
	}

	tempto = strdup(argv[0]);

	// Figure out if the recepient is local

	if (local_domain(tempto)) {
		if (debug) {
			syslog(LOG_INFO, "Debug: Adding %s to local recipients.", argv[0]);
		}

		priv->mlfi_to = (char **) realloc((char **) priv->mlfi_to, sizeof(char *) * (priv->mlfi_to_count + 1));

		if (priv->mlfi_to == NULL) {
			syslog(LOG_ERR, "Error: Failed to malloc()/realloc() recipient recipient array in mlfi_envrcpt()");
			return SMFIS_TEMPFAIL;
		}

		priv->mlfi_to[priv->mlfi_to_count] = strdup(argv[0]);

		if (priv->mlfi_to[priv->mlfi_to_count] == NULL) {
			syslog(LOG_ERR, "Error: Failed to strdup() recipient address in mlfi_envrcpt()");
			return SMFIS_TEMPFAIL;
		}

		// Increment the number of recipients
		priv->mlfi_to_count++;
	}
	
	freemem(tempto, "tempto in mlfi_envrcpt");
	
	return SMFIS_CONTINUE;
} /* mlfi_envrcpt */


/******************************* mlfi_header **********************************/
sfsistat
mlfi_header(ctx, headerf, headerv)
	SMFICTX *ctx;
	char *headerf;
	unsigned char *headerv;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);

	/* Prevent looping of vacation messages */

	if (!strncasecmp(headerf, "X-Loop", 6)) {
		if (!strncasecmp(headerv, "auto-response", 13)) {
			/* Stop milter processing and accept message */
			(void) mlfi_message_cleanup(ctx);
			return SMFIS_ACCEPT;
		} else {
			return SMFIS_CONTINUE;
		}
	} else if (!strncasecmp(headerf, "Precedence", 10)) {
		if (!strncasecmp(headerv, "bulk", 4)) {
			/* Mark message as a bulk message */
			priv->bulk = 1;
		}
	} else if (!strncasecmp(headerf, "X-MailScanner-SpamScore", 23)) {
		if (strlen(headerv) >= 3) {
			/* Mark message as a spam message */
			priv->spam = 1;
		}
	} else if (!strncasecmp(headerf, "X-Spam-Flag", 11)) {
		if (!strncasecmp(headerv, "yes", 3)) {
			/* Mark message as a spam message */
			priv->spam = 1;
		}
	} else if (!strncasecmp(headerf, "Subject", 7)) {
		priv->subject = strdup(headerv);
		if (debug) {
			syslog(LOG_INFO, "Debug: Subject: %s", priv->subject);
		}
	}

	return SMFIS_CONTINUE;
}


/******************************** mlfi_eom ************************************/
sfsistat
mlfi_eom(ctx)
	SMFICTX *ctx;
{
	MlfiPriv *priv = (MlfiPriv *) smfi_getpriv(ctx);
	int i,j;										/* Temporary ints */
	char *username = NULL;			/* Username of the person who the e-mail is for */
	
	/* vacation message handling */
	vacation_msg_list_t vmsg_list = (vacation_msg_list_t) NULL;
	vacation_msg_t vmsg = (vacation_msg_t) NULL;

	/* For each local recepient, do some checks */
	for (i = 0; i < priv->mlfi_to_count; i++) {
		/* Get the username from the To: header */
		username = extract_account(priv->mlfi_to[i]);

		if ((username == NULL) || !strlen(username)) {
			syslog(LOG_ERR, "Error: Could not extract username from '%s'", priv->mlfi_to[i]);
			(void) mlfi_message_cleanup(ctx);
			return SMFIS_TEMPFAIL;
		}

		if (debug) {
			syslog(LOG_INFO, "Debug: Local user: %s", priv->mlfi_to[i]);
		}

		if (is_blacklisted(priv->mlfi_from)) {
			smfi_addheader(ctx, "X-AutoRespond", "NO, sender blacklisted");
		} else if (priv->spam) {
			smfi_addheader(ctx, "X-AutoRespond", "NO, spam mail");
		} else if (priv->bulk) {
			smfi_addheader(ctx, "X-AutoRespond", "NO, bulk mail");
		} else {

			if (debug) {
				syslog(LOG_INFO, "Debug: Checking for vacation message");
			}

			long fl = 0; //file length
			FILE *fp = NULL;
			char msg_file[MSG_FILE_BUFLEN];

			{
				memset(msg_file, 0, MSG_FILE_BUFLEN);
				snprintf(msg_file, MSG_FILE_BUFLEN - 1, "%s/%s", config.msg_path, username);
	
				if (debug) {
					syslog(LOG_INFO, "Debug: Searching file: %s", msg_file);
				}

				fp = fopen(msg_file, "r");
			}

			if (fp == NULL) {
				freemem(username, "username in mlfi_eom, buffer malloc failed");
				username = extract_domainname(priv->mlfi_to[i]);

				memset(msg_file, 0, MSG_FILE_BUFLEN);
				snprintf(msg_file, MSG_FILE_BUFLEN - 1, "%s/%s", config.msg_path, username);
	
				if (debug) {
					syslog(LOG_INFO, "Debug: Searching file: %s", msg_file);
				}

				fp = fopen(msg_file, "r");
			}
			
			if (fp != NULL) {
					// obtain file size:
					fseek(fp, 0 , SEEK_END);
					fl = ftell(fp);
					rewind(fp);
			}
			
			if (fl != 0) { //only if the file is not empty (empty files are used to avoid automatic catch-all functionality)
				
				if (debug) {
					syslog(LOG_INFO, "Debug: Found vacation message");
				}
		
				/* 
					 Vacation message queuing logic: for any given message, 1 or more of the recipients
					 might have a vacation message.	We will queue each formatted message as an element
					 on vmsg_list, and send each messages at the end of smfi_eom, if we make it that far... */
					 
				/* Note vmsg_list will remain NULL unless at least 1 recipient has a vacation message */
				if(!vmsg_list) {
					vmsg_list = create_vacation_msg_list();
				}
				vmsg = create_vacation_msg();
				append_vacation_msg(vmsg_list, vmsg);

	
				{
					/* Basic "From:" header */
					append_to_vacation_msg(vmsg, "From: ");
					append_to_vacation_msg(vmsg, priv->mlfi_to[i]);
					append_to_vacation_msg(vmsg, "\r\n");
				}
				
				/* Get value of vacation subject */
				if (fgets(msg_file, MAX_DOMAIN_LENGTH, fp) != NULL)
				{
					append_to_vacation_msg(vmsg, "Subject: ");
					append_to_vacation_msg(vmsg, msg_file);
					//append_to_vacation_msg(vmsg, "\r\n");

				} else if (priv->subject != NULL) {
					snprintf(msg_file, MSG_FILE_BUFLEN - 1, "AutoRespond: %s", priv->subject);
					append_to_vacation_msg(vmsg, "Subject: ");
					append_to_vacation_msg(vmsg, msg_file);
					append_to_vacation_msg(vmsg, "\r\n");

				} else {
					append_to_vacation_msg(vmsg, "Subject: ");
					append_to_vacation_msg(vmsg, "AutoRespond: [no subject]");
					append_to_vacation_msg(vmsg, "\r\n");
				}
				
				append_to_vacation_msg(vmsg, "To: ");
				append_to_vacation_msg(vmsg, priv->mlfi_from);
				append_to_vacation_msg(vmsg, "\r\n");
				append_to_vacation_msg(vmsg, "X-Loop: auto-response\r\n");
				append_to_vacation_msg(vmsg, "\r\n"); /* we need this! */
			

				{
					char * pos;
					while (fgets(msg_file, MAX_DOMAIN_LENGTH, fp) != NULL) {
						j = strlen(msg_file);
						while (j>0)
						{
							j--;
			
							if (msg_file[j] == '\r' || msg_file[j] == '\n')
								msg_file[j] = '\0';
							else
								j = 0; //exit
						}

#define SEARCH_STRING "%subject"
#define SEARCH_LENGTH 8
						
						pos = strstr(msg_file,SEARCH_STRING);
						if(pos!=NULL)
						{
							freemem(username, "username in mlfi_eom, buffer malloc failed");

							int bufsize = strlen(msg_file) - SEARCH_LENGTH + strlen(priv->subject);
							username = (char *) malloc (sizeof(char) * (bufsize + 1));

							int size = pos - msg_file;
							strncpy(username, msg_file, size);
							strncpy(username + size, priv->subject, strlen(priv->subject));
							strncpy(username + size + strlen(priv->subject), msg_file + size + SEARCH_LENGTH, strlen(msg_file) - size - SEARCH_LENGTH);

							username[bufsize] = '\0';
							append_to_vacation_msg(vmsg, username);
						}
						else
						{
							append_to_vacation_msg(vmsg, msg_file);
						}
						append_to_vacation_msg(vmsg, "\r\n");
					}
				}
				
				smfi_addheader(ctx, "X-AutoRespond", "YES");
				
				if (debug) {
					syslog(LOG_INFO, "Debug: Queued vacation reply. From: %s To: %s", priv->mlfi_to[i], priv->mlfi_from);
				}

			}
			else
			{
				smfi_addheader(ctx, "X-AutoRespond", "NO, not activated");
			}

			if (fp != NULL) fclose(fp);

		}
	
		if (debug) {
			syslog(LOG_INFO, "Debug: Done processing %s", username);
		}

		freemem(username, "username in mlfi_eom, username cleanup");
	}
	
	/* send all queued vacation messages */
	if(vmsg_list) {

		if (debug) {
		syslog(LOG_INFO, "Debug: Sending vacation message replies");
		}	

		send_vacation_messages(config.sendmail, vmsg_list, debug);
		delete_vacation_msg_list(vmsg_list);
		vmsg_list = (vacation_msg_list_t) NULL;
	}

	return mlfi_message_cleanup(ctx);
} /* mlfi_eom */


/******************************** mlfi_abort **********************************/
sfsistat
mlfi_abort(ctx)
	SMFICTX *ctx;
{
	if (debug) {
		syslog(LOG_INFO, "Debug: Aborting connection!");
	}

	return mlfi_message_cleanup(ctx);
} /* mlfi_abort */


/********************************* mlfi_close *********************************/
sfsistat
mlfi_close(ctx)
	SMFICTX *ctx;
{
	if (debug) {
		syslog(LOG_INFO, "Debug: Closing Connection!");
	}

	return mlfi_connection_cleanup(ctx);
} /* mlfi_close */


/**************************** smfilter callbacks ******************************/
struct smfiDesc smfilter =
{
	"fv_milter",	/* filter name */
	SMFI_VERSION, /* version code -- do not change */
	SMFIF_ADDHDRS | SMFIF_CHGHDRS | SMFIF_ADDRCPT | SMFIF_DELRCPT, /* flags */
	mlfi_connect, /* connection info filter */
	NULL,				 /* SMTP HELO command filter */
	mlfi_envfrom, /* envelope sender filter */
	mlfi_envrcpt, /* envelope recipient filter */
	mlfi_header,	/* header filter */
	NULL,				 /* end of header */
	NULL,				 /* body block filter */
	mlfi_eom,		 /* end of message */
	mlfi_abort,	 /* message aborted */
	mlfi_close,	 /* connection cleanup */
};


/*********************************** main *************************************/
int main (argc, argv)
	int argc;
	char *argv[];
{
	int c,											 /* getopt() return value */
			i,											 /* Domains array index */
			status;									 /* smfi_main return status */
	const char *args = "vdc:p:"; /* Valid command line arguments */
	extern char *optarg;				 /* Given command line argument */

	if (argc < 2) {
		usage(argv[0]);
		exit(EX_USAGE);
	}

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != -1) {
		switch (c) {
			case 'p':
				if (optarg == NULL || *optarg == '\0') {
					(void) fprintf(stderr, "Illegal connection: %s\n", optarg);
					exit(EX_USAGE);
				}

				if (strncasecmp(optarg, "unix:", 5) == 0) {
					unlink(optarg + 5);
				} else if (strncasecmp(optarg, "local:", 6) == 0) {
					unlink(optarg + 6);
				} else {
					usage(argv[0]);
					exit(EX_USAGE);
				}

				if (smfi_setconn(optarg) == MI_FAILURE) {
					(void) fprintf(stderr, "smfi_setconn failed\n");
					exit(EX_SOFTWARE);
				}

				break;

			case 'c':
				if( optarg == NULL ) {
					(void) fprintf( stderr, "The -c option requires a file name\n" );
					exit( EX_USAGE );
				} else {
					config_filename = optarg;
				}
				
				break;
				
			case 'd':
				fprintf(stderr, "\nDebugging turned on. DO NOT run in production with this option!\n\n");
				debug = 1;
				break;

			case 'v':
				fprintf(stderr, "fv_milter %s\n", VERSION);
				exit(EX_USAGE);

			default:
				usage(argv[0]);
				exit(EX_USAGE);
		}
	}
	
	srand(time(NULL));
	
	/* open syslog */
	openlog("fv_milter", LOG_PID, LOG_MAIL );

	if (debug) {
		syslog(LOG_INFO, "Started in debug mode.");
	} else {
		syslog(LOG_INFO, "Started.");
	}

	/* Register our callbacks */
	if (smfi_register(smfilter) == MI_FAILURE) {
		fprintf(stderr, "Call to smfi_register() failed\n");
		exit(EX_UNAVAILABLE);
	}

	/* Read config file */
	if (read_config()) {
		/* Read local domains */
		if (!read_domains(config.domains_file)) {
			fprintf(stderr, "Domain read failed\n");
		} else if (config.blacklist_file!=NULL && !read_blacklist(config.blacklist_file)) {
			fprintf(stderr, "Blacklist read failed\n");
		} else {

			/* Give control to libmilter */
			if ((status = smfi_main()) == MI_FAILURE) {
				fprintf(stderr, "Call to smfi_main() failed\n");
				exit(EX_UNAVAILABLE);
			}

		}
	} else {
		fprintf(stderr, "Config read failed\n");
	}

	/* Clean up if exiting */
	if (domains != NULL) {
		for (i = 0; i < num_domains; i++) {
			freemem(domains[i], "domain[i]");
		}

		freemem(domains, "domains");
	}
	
	/* Clean up config data */
	if (config.sendmail != NULL) {
		freemem(config.sendmail, "config.sendmail");
	}

	if (config.domains_file != NULL) {
		freemem(config.domains_file, "config.domains_file");
	}

	if (config.blacklist_file != NULL) {
		freemem(config.blacklist_file, "config.blacklist_file");
	}

	if (config.msg_path != NULL) {
		freemem(config.msg_path, "config.msg_path");
	}

	syslog(LOG_INFO, "Exiting.");

	return status;
} /* main */
